import asyncio
import os
import shutil
from collections.abc import AsyncIterator, Callable
from contextlib import AbstractContextManager
from typing import Any, Literal

import orjson
import structlog
import zmq
import zmq.asyncio
from langchain_core.runnables.config import RunnableConfig
from langchain_core.runnables.graph import Edge, Node
from langchain_core.runnables.graph import Graph as DrawableGraph
from langchain_core.runnables.schema import (
    CustomStreamEvent,
    StandardStreamEvent,
    StreamEvent,
)
from langgraph.checkpoint.base.id import uuid6
from langgraph.checkpoint.serde.base import SerializerProtocol
from langgraph.pregel.types import PregelTask, StateSnapshot
from langgraph.store.base import GetOp, Item, ListNamespacesOp, PutOp, SearchOp
from langgraph.types import Command, Interrupt
from pydantic import BaseModel
from starlette.exceptions import HTTPException
from zmq.utils.monitor import recv_monitor_message

from langgraph_api.js.base import BaseRemotePregel
from langgraph_api.js.errors import RemoteException
from langgraph_api.js.schema import (
    ErrorData,
    RequestPayload,
    ResponsePayload,
    StreamData,
)
from langgraph_api.serde import json_dumpb, json_loads
from langgraph_api.utils import AsyncConnectionProto

logger = structlog.stdlib.get_logger(__name__)

CLIENT_ADDR = "tcp://0.0.0.0:5556"
REMOTE_ADDR = "tcp://0.0.0.0:5555"
MONITOR_ADDR = "inproc://router.remote"

WAIT_FOR_REQUEST_TTL_SECONDS = 15.0

context = zmq.asyncio.Context()
clientDealer = context.socket(zmq.DEALER)
remoteRouter = context.socket(zmq.ROUTER)

remoteRouter.monitor(
    MONITOR_ADDR,
    zmq.EVENT_HANDSHAKE_SUCCEEDED | zmq.EVENT_DISCONNECTED,
)
remoteMonitor = context.socket(zmq.PAIR)
remoteMonitor.setsockopt(zmq.LINGER, 0)

REMOTE_MGS_MAP: dict[str, asyncio.Queue] = {}


class HeartbeatPing:
    pass


class RequestTimeout(Exception):
    method: str

    def __init__(self, method: str):
        self.method = method

    def __str__(self):
        return f'Request to "{self.method}" timed out'


class RequestContext(AbstractContextManager["RequestContext"]):
    id: str
    method: str

    def __init__(self, method: str):
        self.id = uuid6().hex
        self.method = method

    def __enter__(self) -> "RequestContext":
        REMOTE_MGS_MAP[self.id] = asyncio.Queue()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        REMOTE_MGS_MAP.pop(self.id, None)

    async def get(self) -> Any:
        try:
            value = await asyncio.wait_for(
                REMOTE_MGS_MAP[self.id].get(), timeout=WAIT_FOR_REQUEST_TTL_SECONDS
            )
        except TimeoutError as exc:
            raise RequestTimeout(self.method) from exc

        if isinstance(value, HeartbeatPing):
            return await self.get()
        if isinstance(value, Exception):
            raise value
        return value


async def _client_stream(method: str, data: Any):
    with RequestContext(method) as context:
        await clientDealer.send(
            orjson.dumps({"method": method, "id": context.id, "data": data})
        )

        while True:
            response: StreamData = await context.get()

            if response["done"]:
                if "value" in response:
                    yield response["value"]
                break

            yield response["value"]


async def _client_invoke(method: str, data: Any):
    with RequestContext(method) as context:
        await clientDealer.send(
            orjson.dumps({"method": method, "id": context.id, "data": data})
        )

        return await context.get()


class RemotePregel(BaseRemotePregel):
    @staticmethod
    def load(graph_id: str):
        model = RemotePregel()
        model.graph_id = graph_id
        return model

    async def astream_events(
        self,
        input: Any,
        config: RunnableConfig | None = None,
        *,
        version: Literal["v1", "v2"],
        **kwargs: Any,
    ) -> AsyncIterator[StreamEvent]:
        if version != "v2":
            raise ValueError("Only v2 of astream_events is supported")

        data = {
            "graph_id": self.graph_id,
            "command" if isinstance(input, Command) else "input": input,
            "config": config,
            **kwargs,
        }

        async for event in _client_stream("streamEvents", data):
            if event["event"] == "on_custom_event":
                yield CustomStreamEvent(**event)
            else:
                yield StandardStreamEvent(**event)

    async def fetch_state_schema(self):
        return await _client_invoke("getSchema", {"graph_id": self.graph_id})

    async def fetch_graph(
        self,
        config: RunnableConfig | None = None,
        *,
        xray: int | bool = False,
    ) -> DrawableGraph:
        response = await _client_invoke(
            "getGraph", {"graph_id": self.graph_id, "config": config, "xray": xray}
        )

        nodes: list[Any] = response.pop("nodes")
        edges: list[Any] = response.pop("edges")

        class NoopModel(BaseModel):
            pass

        return DrawableGraph(
            {
                data["id"]: Node(
                    data["id"], data["id"], NoopModel(), data.get("metadata")
                )
                for data in nodes
            },
            {
                Edge(
                    data["source"],
                    data["target"],
                    data.get("data"),
                    data.get("conditional", False),
                )
                for data in edges
            },
        )

    async def fetch_subgraphs(
        self, *, namespace: str | None = None, recurse: bool = False
    ) -> dict[str, dict]:
        return await _client_invoke(
            "getSubgraphs",
            {"graph_id": self.graph_id, "namespace": namespace, "recurse": recurse},
        )

    def _convert_state_snapshot(self, item: dict) -> StateSnapshot:
        def _convert_tasks(tasks: list[dict]) -> tuple[PregelTask, ...]:
            result: list[PregelTask] = []
            for task in tasks:
                state = task.get("state")

                if state and isinstance(state, dict) and "config" in state:
                    state = self._convert_state_snapshot(state)

                result.append(
                    PregelTask(
                        task["id"],
                        task["name"],
                        tuple(task["path"]) if task.get("path") else tuple(),
                        # TODO: figure out how to properly deserialise errors
                        task.get("error"),
                        (
                            tuple(
                                Interrupt(
                                    value=interrupt["value"],
                                    when=interrupt["when"],
                                    resumable=interrupt.get("resumable", True),
                                    ns=interrupt.get("ns"),
                                )
                                for interrupt in task.get("interrupts")
                            )
                            if task.get("interrupts")
                            else []
                        ),
                        state,
                    )
                )
            return tuple(result)

        return StateSnapshot(
            item.get("values"),
            item.get("next"),
            item.get("config"),
            item.get("metadata"),
            item.get("createdAt"),
            item.get("parentConfig"),
            _convert_tasks(item.get("tasks", [])),
        )

    async def aget_state(
        self, config: RunnableConfig, *, subgraphs: bool = False
    ) -> StateSnapshot:
        return self._convert_state_snapshot(
            await _client_invoke(
                "getState",
                {"graph_id": self.graph_id, "config": config, "subgraphs": subgraphs},
            )
        )

    async def aupdate_state(
        self,
        config: RunnableConfig,
        values: dict[str, Any] | Any,
        as_node: str | None = None,
    ) -> RunnableConfig:
        response = await _client_invoke(
            "updateState",
            {
                "graph_id": self.graph_id,
                "config": config,
                "values": values,
                "as_node": as_node,
            },
        )
        return RunnableConfig(**response)

    async def aget_state_history(
        self,
        config: RunnableConfig,
        *,
        filter: dict[str, Any] | None = None,
        before: RunnableConfig | None = None,
        limit: int | None = None,
    ) -> AsyncIterator[StateSnapshot]:
        async for event in _client_stream(
            "getStateHistory",
            {
                "graph_id": self.graph_id,
                "config": config,
                "limit": limit,
                "filter": filter,
                "before": before,
            },
        ):
            yield self._convert_state_snapshot(event)

    def get_graph(
        self,
        config: RunnableConfig | None = None,
        *,
        xray: int | bool = False,
    ) -> dict[str, Any]:
        raise Exception("Not implemented")

    def get_input_schema(self, config: RunnableConfig | None = None) -> type[BaseModel]:
        raise Exception("Not implemented")

    def get_output_schema(
        self, config: RunnableConfig | None = None
    ) -> type[BaseModel]:
        raise Exception("Not implemented")

    def config_schema(self) -> type[BaseModel]:
        raise Exception("Not implemented")

    async def invoke(self, input: Any, config: RunnableConfig | None = None):
        raise Exception("Not implemented")


async def run_js_process(paths_str: str, watch: bool = False):
    # check if tsx is available
    tsx_path = shutil.which("tsx")
    if tsx_path is None:
        raise FileNotFoundError("tsx not found in PATH")
    attempt = 0
    while not asyncio.current_task().cancelled():
        client_file = os.path.join(os.path.dirname(__file__), "client.new.mts")
        args = ("tsx", client_file)
        if watch:
            args = ("tsx", "watch", client_file, "--skip-schema-cache")
        try:
            process = await asyncio.create_subprocess_exec(
                *args,
                env={
                    "LANGSERVE_GRAPHS": paths_str,
                    "LANGCHAIN_CALLBACKS_BACKGROUND": "true",
                    "CHOKIDAR_USEPOLLING": "true",
                    **os.environ,
                },
            )
            code = await process.wait()
            raise Exception(f"JS process exited with code {code}")
        except asyncio.CancelledError:
            logger.info("Terminating JS graphs process")
            try:
                process.terminate()
                await process.wait()
            except (UnboundLocalError, ProcessLookupError):
                pass
            raise
        except Exception:
            if attempt >= 3:
                raise
            else:
                logger.warning(f"Retrying JS process {3 - attempt} more times...")
                attempt += 1


def _get_passthrough_checkpointer(conn: AsyncConnectionProto):
    from langgraph_storage.checkpoint import Checkpointer

    class PassthroughSerialiser(SerializerProtocol):
        def dumps(self, obj: Any) -> bytes:
            return json_dumpb(obj)

        def dumps_typed(self, obj: Any) -> tuple[str, bytes]:
            return "json", json_dumpb(obj)

        def loads(self, data: bytes) -> Any:
            return orjson.loads(data)

        def loads_typed(self, data: tuple[str, bytes]) -> Any:
            type, payload = data
            if type != "json":
                raise ValueError(f"Unsupported type {type}")
            return orjson.loads(payload)

    checkpointer = Checkpointer(conn)

    # This checkpointer does not attempt to revive LC-objects.
    # Instead, it will pass through the JSON values as-is.
    checkpointer.serde = PassthroughSerialiser()

    return checkpointer


def _get_passthrough_store():
    from langgraph_storage.store import Store

    return Store()


async def run_remote_checkpointer():
    from langgraph_storage.database import connect

    async def checkpointer_list(payload: dict):
        """Search checkpoints"""

        result = []
        async with connect() as conn:
            checkpointer = _get_passthrough_checkpointer(conn)
            async for item in checkpointer.alist(
                config=payload.get("config"),
                limit=payload.get("limit"),
                before=payload.get("before"),
                filter=payload.get("filter"),
            ):
                result.append(item)

        return result

    async def checkpointer_put(payload: dict):
        """Put the new checkpoint metadata"""

        async with connect() as conn:
            checkpointer = _get_passthrough_checkpointer(conn)
            return await checkpointer.aput(
                payload["config"],
                payload["checkpoint"],
                payload["metadata"],
                payload.get("new_versions", {}),
            )

    async def checkpointer_get_tuple(payload: dict):
        """Get actual checkpoint values (reads)"""

        async with connect() as conn:
            checkpointer = _get_passthrough_checkpointer(conn)
            return await checkpointer.aget_tuple(config=payload["config"])

    async def checkpointer_put_writes(payload: dict):
        """Put actual checkpoint values (writes)"""

        async with connect() as conn:
            checkpointer = _get_passthrough_checkpointer(conn)
            return await checkpointer.aput_writes(
                payload["config"],
                payload["writes"],
                payload["taskId"],
            )

    async def store_batch(payload: dict):
        """Batch operations on the store"""
        operations = payload.get("operations", [])

        if not operations:
            raise ValueError("No operations provided")

        # Convert raw operations to proper objects
        processed_operations = []
        for op in operations:
            if "value" in op:
                processed_operations.append(
                    PutOp(
                        namespace=tuple(op["namespace"]),
                        key=op["key"],
                        value=op["value"],
                    )
                )
            elif "namespace_prefix" in op:
                processed_operations.append(
                    SearchOp(
                        namespace_prefix=tuple(op["namespace_prefix"]),
                        filter=op.get("filter"),
                        limit=op.get("limit", 10),
                        offset=op.get("offset", 0),
                    )
                )

            elif "namespace" in op and "key" in op:
                processed_operations.append(
                    GetOp(namespace=tuple(op["namespace"]), key=op["key"])
                )
            elif "match_conditions" in op:
                processed_operations.append(
                    ListNamespacesOp(
                        match_conditions=tuple(op["match_conditions"]),
                        max_depth=op.get("max_depth"),
                        limit=op.get("limit", 100),
                        offset=op.get("offset", 0),
                    )
                )
            else:
                raise ValueError(f"Unknown operation type: {op}")

        store = _get_passthrough_store()
        results = await store.abatch(processed_operations)

        # Handle potentially undefined or non-dict results
        processed_results = []
        # Result is of type: Union[Item, list[Item], list[tuple[str, ...]], None]
        for result in results:
            if isinstance(result, Item):
                processed_results.append(result.dict())
            elif isinstance(result, dict):
                processed_results.append(result)
            elif isinstance(result, list):
                coerced = []
                for res in result:
                    if isinstance(res, Item):
                        coerced.append(res.dict())
                    elif isinstance(res, tuple):
                        coerced.append(list(res))
                    elif res is None:
                        coerced.append(res)
                    else:
                        coerced.append(str(res))
                processed_results.append(coerced)
            elif result is None:
                processed_results.append(None)
            else:
                processed_results.append(str(result))
        return processed_results

    async def store_get(payload: dict):
        """Get store data"""
        namespaces_str = payload.get("namespaces")
        key = payload.get("key")

        if not namespaces_str or not key:
            raise ValueError("Both namespaces and key are required")

        namespaces = namespaces_str.split(".")

        store = _get_passthrough_store()
        result = await store.aget(namespaces, key)

        return result

    async def store_put(payload: dict):
        """Put the new store data"""

        namespace = tuple(payload["namespace"].split("."))
        key = payload["key"]
        value = payload["value"]
        index = payload.get("index")

        store = _get_passthrough_store()
        await store.aput(namespace, key, value, index=index)

        return {"success": True}

    async def store_search(payload: dict):
        """Search stores"""
        namespace_prefix = tuple(payload["namespace_prefix"])
        filter = payload.get("filter")
        limit = payload.get("limit", 10)
        offset = payload.get("offset", 0)
        query = payload.get("query")

        store = _get_passthrough_store()
        result = await store.asearch(
            namespace_prefix, filter=filter, limit=limit, offset=offset, query=query
        )

        return [item.dict() for item in result]

    async def store_delete(payload: dict):
        """Delete store data"""

        namespace = tuple(payload["namespace"])
        key = payload["key"]

        store = _get_passthrough_store()
        await store.adelete(namespace, key)

        return {"success": True}

    async def store_list_namespaces(payload: dict):
        """List all namespaces"""
        prefix = tuple(payload.get("prefix", [])) or None
        suffix = tuple(payload.get("suffix", [])) or None
        max_depth = payload.get("max_depth")
        limit = payload.get("limit", 100)
        offset = payload.get("offset", 0)

        store = _get_passthrough_store()
        result = await store.alist_namespaces(
            prefix=prefix,
            suffix=suffix,
            max_depth=max_depth,
            limit=limit,
            offset=offset,
        )

        return [list(ns) for ns in result]

    methods = {
        checkpointer_get_tuple,
        checkpointer_list,
        checkpointer_put,
        checkpointer_put_writes,
        store_get,
        store_put,
        store_delete,
        store_search,
        store_batch,
        store_list_namespaces,
    }

    method_map: dict[str, Callable[[dict], Any]] = {
        method.__name__: method for method in methods
    }

    with (
        clientDealer.connect(CLIENT_ADDR),
        remoteMonitor.connect(MONITOR_ADDR),
        remoteRouter.bind(REMOTE_ADDR),
    ):
        poller = zmq.asyncio.Poller()
        poller.register(remoteRouter, zmq.POLLIN)
        poller.register(clientDealer, zmq.POLLIN)
        poller.register(remoteMonitor, zmq.POLLIN)

        while not asyncio.current_task().cancelled():
            events = dict(await poller.poll())

            if remoteRouter in events:
                identity, raw_req = await remoteRouter.recv_multipart()
                req: RequestPayload = json_loads(raw_req)

                method = req.get("method")
                id = req.get("id")

                try:
                    if fn := method_map.get(method):
                        data = await fn(req.get("data"))
                    else:
                        raise ValueError(f"Unknown method {method}")

                    resp = {"method": method, "id": id, "success": True, "data": data}
                    await remoteRouter.send_multipart([identity, json_dumpb(resp)])

                except BaseException as exc:
                    await logger.aexception(
                        f"Error in remote method {method}", exc_info=exc
                    )

                    resp = {"method": method, "id": id, "success": False, "data": exc}
                    await remoteRouter.send_multipart([identity, json_dumpb(resp)])

            if clientDealer in events:
                response: ResponsePayload = json_loads(await clientDealer.recv())
                queue = REMOTE_MGS_MAP.get(response["id"])

                if queue:
                    if response.get("success") is None:
                        await queue.put(HeartbeatPing())

                    elif response["success"]:
                        data: Any = response["data"]
                        await queue.put(data)
                    else:
                        data: ErrorData = response["data"]
                        await queue.put(RemoteException(data["error"], data["message"]))

            if remoteMonitor in events:
                msg = await recv_monitor_message(remoteMonitor)
                if msg["event"] == zmq.EVENT_HANDSHAKE_SUCCEEDED:
                    logger.info("JS worker connected")
                elif msg["event"] == zmq.EVENT_DISCONNECTED:
                    logger.info("JS worker disconnected")


async def wait_until_js_ready():
    attempt = 0
    while not asyncio.current_task().cancelled():
        try:
            await _client_invoke("ok", {})
            return
        except (RemoteException, TimeoutError, zmq.error.ZMQBaseError):
            if attempt > 240:
                raise
            else:
                attempt += 1
                await asyncio.sleep(0.5)


async def js_healthcheck():
    try:
        await _client_invoke("ok", {})
        return True
    except (RemoteException, TimeoutError, zmq.error.ZMQBaseError) as exc:
        logger.warning(
            "JS healthcheck failed. Either the JS server is not running or the event loop is blocked by a CPU-intensive task.",
            error=exc,
        )
        raise HTTPException(
            status_code=500,
            detail="JS healthcheck failed. Either the JS server is not running or the event loop is blocked by a CPU-intensive task.",
        ) from exc
